/* Bit Manipulation Functions */

#include <string.h> /* for strlen() or strcpy() or strcmp() */

#include "./types.h"
#include "./bits.h"

/** assign a bit to targetByte, and return targetByte **/
uchar putBit(uchar targetByte, uint bit, int bitToReplace)
{
  int i=0;
  uchar targetBit=0;
  uchar applicationBit=0;

  /* get the bit we are going to change */
  targetBit=targetByte;
  targetBit <<= (8-bitToReplace);  /* get rid of bits on the left */
  targetBit >>= (7);               /* get rid of bits on the right */
  targetBit <<= (bitToReplace-1);  /* move the bit back to where it belongs */

  /* fix up our application bit */
  applicationBit=1;                    /* application bit must be 1 */
  applicationBit <<= (bitToReplace-1); /* make sure the application bit is in the correct location */

  /* apply the application bit, if neccessary. we only need to modify the bit if its not already the same as what we are tryign to make it */
  if (bit == 1 && targetBit == 0) /* if the bit should be 1, but is 0... */
  {
    targetByte += applicationBit; /* the bit is 0, so add 1 to make it 1 */
  }else if (bit == 0 && targetBit != 0){
    targetByte -= applicationBit; /* the bit is 1, so subtract 1 to make it 0*/
  }

  return targetByte;
}

/** retrieve a bit from sourceByte and return it **/
uint getBit(uchar sourceByte, uint bitToGet)
{
  uint bit=0;

  sourceByte <<= (8-bitToGet); /* remove the extra bits on the left */
  sourceByte >>= 7;            /* remove the extra bits on the right */

  if (sourceByte != 0) /* if its not 0, then its gotta be 1 */
    bit=1; /* just incase its 2, or 4, or 8, ... */

  return bit;
}

/* flip bit in sourceByte and return sourceByte */
uchar flipBit(uchar byte, uint bitToFlip)
{
  if (getBit(byte, bitToFlip) == 0) /* if the bit is 0 */
  {
    byte=putBit(byte, 1, bitToFlip); /* you get the idea.... */
  }else{
    byte=putBit(byte, 0, bitToFlip);
  }

  return byte;
}

/* reverse the order of bits c thgough b in a byte */
uchar reverseBits(uchar byte, uint b, uint c)
{
  for (; b < c; c--, b++) /* if there are no more sets of bits between b and c */
    byte=swapBits(byte, b, c);

  return byte;
}

/* flip all the bits in a byte */
uchar flipByte(uchar byte)
{
  uint i;

  /* flip 'em all */ 
  for (i=8; i >=1; i--)
    flipBit(byte, i);

  return byte;
}

/* swap the values of two bits within byte */
uchar swapBits(uchar byte, uint a, uint b)
{
  uchar copy;

  copy=byte; /* for the same comcept as trying to swap two values without a temp...using a temp is the easiest route */

  copy=putBit(copy, getBit(byte, a), b); /* b goes in place of c */
  copy=putBit(copy, getBit(byte, b), a); /* and vice versa */
}

/** Char to Binary.
    convert a char to binary, and put it into *str,
    provide your own string at least 9 chars long! **/
void char2bin(uchar a, char *str)
{
  uint i, b, len;

  len = strlen(str);

// for some reason this just doesnt work
//  /* make sure the string is long enough */
//  if (len < 8)
//    str=realloc(&str, sizeof(char) * 9-len); /* the length will be 9, the last for a null terminator. */

  for (i=8, b=0; i>=1; i--,b++)
  {
    if (getBit(a, i) == 0)
      str[b]='0';
    else
      str[b]='1';
  }
  str[8] = '\0'; /* add a peso! */

  return;
}
