// note0: I suggest you turn word wrapping off :-)
// note1: this program is invoked with any number of options with one argument each in this manner: shade -option argument_to_option
// note2: a bit's first bit is the one farthest right, and its 8th is farthest left (ie 87654321)
// note3: throughout the source, the terms 'options', 'switches', and 'parameters' are used interchangably
// note4: a switch is passed with arguments such that "-r" is a switch and "host" is an ARGUMENT when given "shade -r host"
// note5: some options can be used to read and write. for ease of documentation, instead of saying "read/write", we'll say "performs"
// note6: when outputing messages, add linebreaks to break lines into segments smaller than, but as close as possible, to 80 chars. ruler:
/*--------------------this--is--a--ruler--80--chars--long---------------------*/

#include <stdio.h>
#include <stdlib.h> /* for atoi() or malloc() */
#include <string.h> /* for strlen() or strcpy() or strcmp() */

#include "./types.h"    /* custom data types */
#include "./files.h"   /* file functions */
#include "./strint.h"   /* string and integer functions */
#include "./bits.h"     /* bitwise functions */
#include "./options.h"  /* command-line agument handling */
#include "./shade.h"    /* function prototypes */

/** think of these #defines as inline functions **/
#define writeHostByte(); {if (output == false) { printf("%c", hostByte); }else{ fwrite(&hostByte, sizeof(uchar), 1, out); }}  /* print a byte, or to a file if one was given */
#define writePayByte(); {if (output == false) { printf("%c", payByte); }else{ fwrite(&payByte, sizeof(uchar), 1, out); }}     /* print a byte, or to a file if one was given */

/* no reason not to be global, these are all available until program terminates anyhow. */
FILE           *host=NULL, *payload=NULL, *out=NULL;

int main(int args, char *arg[])
{
  uchar  hostByte=0,  //byte read in from host file
         payByte=0,   //byte read in from payload file
         mode=0;      //program mode
  uint   head=0,      //bytes to skip at head of host
         tail=0,      //bytes to skip at tail of host
         len=0,       //length of host file
         i=0, c=0,    //temp integers
         hostBytesPerformed=0, payloadBytesPerformed=0, bitsReadOrWritten=0, bitsPerHostByte=1;
  int    hostBytesToPerform=-1, payloadBytesToPerform=-1, // if it were 0, it would perform none. -1 reads all.
         optIndex=-1,  //array index of command-line option's argument
         skip=0,       //skip this many bytes after performing each byte
         hops=0;       //skip counter
  bool   printChars=true,  //sooner or later we'll implement writing 1's and 0's instead of data
         output=false;     //whether or not an output file was supplied

  extern FILE *host, *payload, *out;
  extern OPTIONS options;

  if (args < 3) /* 3 is the least amount of parameters neccessary: shade -e host */
  { // things commented out here haven't been implemented yet
    printf("Shade\n");
    printf("  In all examples, hfile=host file, pfile=payload file, and ofile=output file.\n");
    printf("  You must first supply a mode:\n");
    printf("    -r [hfile]             Read mode: Extract payload from host.\n");
    printf("    -w [hfile] -p [pfile]  Write mode: Inject payload to host.\n");
    printf("    -e [hfile]             Evaluate Mode: Determine amount of data host can hold.\n");
//    printf("    -f                    Output bits instead of bytes.\n");
    printf("  Options:\n");
    printf("    -m off                 Suppress messages.\n");
    printf("    -o [ofile]             Output processed data to file.\n");
    printf("    -l [#]                 Specify length of payload in bytes.\n");
    printf("    -b [1-8]               Use [1-8] bits per host byte.\n");
    printf("    -B [#]                 Perform only [#] bytes of host.\n");
    printf("    -s [#]                 Skip [#] host bytes after performing each one.\n");
    printf("    -h [#]                 Skip [#] bytes at head of host.\n");
    printf("    -t [#]                 Skip [#] bytes at tail of host.\n");
//  printf("    -P [...]              Xor-crypt payload against repeating bit-sequence.\n");
//  printf("    -x [...]              Xor-crypt payload against repeating passphrase.\n");
//  printf("    -R                    Reverse payload file data (inject/extract backwards).\n");
//  printf("    -P [phrase]           Use payload end-of-data phrase in host.\n");
//  printf("    -a [phrase]           Begin performing after [phrase] is read from host file.\n");
//  printf("    -A [phrase]           Stop performing when [phrase] is read from host file.\n");
    printf("    -f [x]                 Perform using pre-defined file type options. Valid\n");
    printf("                           arguments for [x] are as follows:\n");
    printf("      bmp8                   Windows 8-bit bitmap, 256 colors. (untested)\n");
    printf("                             presets: -h 1078. If option -h is supplied, it's\n");
    printf("                             argument will be added to the preset value of 1078.\n");
    printf("      bmp24                  Windows 24-bit bitmap, 16 million colors.\n");
    printf("                             presets: -h 75 -t 1. If option -h is supplied,\n");
    printf("                             argument will be added to the preset value of 75.\n");
    printf("                             All other options will override the presets.\n");
    printf("  Sequential Arbitrary Bit Manipulation Operations per Host Byte:\n");
    printf("    -R [A][B]              Reverse order of bits [A] to [B]. [A] an [B] must\n");
    printf("                           both be integers between 1 and 8, and [B] must be\n");
    printf("                           larger than [A]\n");
    printf("    -i [[#][#]...[#]]      Invert up to 8 bits. Supplying the same bit multiple\n");
    printf("                           times inverts it accordingly.\n");
    printf("    -T [A][B]              Transpose bits [A] and [B]. [A] and [B] must both be\n");
    printf("                           integers between 1 and 8.\n");
//  printf("    -X [########]          Xor-crypt byte against 8-bit-sequence.\n");

    return(0);
  }	

  /** Put options and arguments into 'opt[]' and 'arg[]' respectively, which are in a global structure variable 'options' defined in options.h **/
  getOptions(args, arg);

  /** Determine Operation Mode (and get index of host file in the passed arguments) **/
  /* check for read mode */
  optIndex=getArg("r");
  if (optIndex != -1)
  {
    mode='r';
    /* open the host file */
    host = fopen(options.arg[optIndex], "rb");

    /* verify that the host file was successfully opened */
    if (host == NULL)
      endProgram(2);
  }else{ /* check for write mode if read mode was not supplied */
    optIndex=getArg("w");
    if (optIndex != -1)
    {
      mode='w';

      /* open the host file */
      host = fopen(options.arg[optIndex], "rb");

      /* verify that the host file was successfully opened */
      if (host == NULL)
        endProgram(3);

      /* open the payload file */
      optIndex=getArg("p");
      if (optIndex != -1)
      {
        /* open the host file */
        payload = fopen(options.arg[optIndex], "rb");

        /* verify that the payload file was successfully opened */
        if (payload == NULL)
          endProgram(4);
      }else{ /* payload was not supplied */
        endProgram(5);
      }

    }else{ /* check for evaluate mode if write or read mode was not supplied */
      optIndex=getArg("e");
      if (optIndex != -1)
      {
        mode='e';
        host = fopen(options.arg[optIndex], "rb"); /* open the host file */

        /* verify that the host file was successfully opened */
        if (host == NULL)
          endProgram(2);
      }else{
        endProgram(6);
      }
    }
  }

  /* get length of host */
  len=(filelen(host));

  /** perform operations pertanent to all program modes **/
  /* are we being asked to output to a file? evaluate mode does not make use of this */
  optIndex=getArg("o");
  if (optIndex != -1)
  {
    output=true;
    /* open the host file */
    out = fopen(options.arg[optIndex], "wb");

    /* verify that the output file was successfully opened */
    if (out == NULL)
      endProgram(7);
  }

  /* are we being asked to skip bytes at the beginning of host? */
  optIndex=getArg("h");
  if (optIndex != -1)
    head=atoi(options.arg[optIndex]); /* we'll need this for later */

  /* are we being asked to process only # host bytes? */
  optIndex=getArg("B");
  if (optIndex != -1)
    hostBytesToPerform=atoi(options.arg[optIndex]);

  /* are we being asked to process only # payload bytes? */
  optIndex=getArg("l");
  if (optIndex != -1)
    payloadBytesToPerform=atoi(options.arg[optIndex]);

  /* are we being asked to output bits instead of chars? */
  optIndex=getArg("f");
  if (optIndex != -1)
    if(options.arg[optIndex][0] == '1')
      printChars=false;

  /* are we being asked to use [1-8] bits per host byte to read/write payload? */
  optIndex=getArg("b");
  if (optIndex != -1)
  {
    bitsPerHostByte=atoi(options.arg[optIndex]);
    if (bitsPerHostByte > 8 || bitsPerHostByte < 1)
      endProgram(8);
  }

  /* are we being asked to apply presets for a specific file type? */
  optIndex=getArg("f");
  if (optIndex != -1)
  {
    if (!strcmp(options.arg[optIndex], "bmp8"))
    {
      head+=1078;
    }
    else if (!strcmp(options.arg[optIndex], "bmp24"))
    {
      head+=75;  /* header is 75 bytes */
      //skip=2;     /* Skip 2 bytes after performing each byte (perform every 3 bytes starting with the first) */

      /* skip one byte at end of file */
      /* we want what ever is shorter: what was passed with -B, or where the tail begins that we're told not to write to by file presets */
      if (((len-head)-1) < hostBytesToPerform)
        hostBytesToPerform=((len-head)-1);
    }
  }

  /* are we being asked to skip bytes at tail of host? this overrides option -B. this should ovverride file type presets. */
  optIndex=getArg("t");
  if (optIndex != -1)
  {
    tail=atoi(options.arg[optIndex]);

    /* we want what ever is shorter: what was passed with -B or -f, or where the tail begins that we're told not to write to */
    if (((len-head)-tail) < hostBytesToPerform)
      hostBytesToPerform=((len-head)-tail);
  }

  /* are we being asked to skip bytes between written bits? this should override file type presets. */
  optIndex=getArg("s");
  if (optIndex != -1)
    skip=atoi(options.arg[optIndex]);

  /* bytes to read must be positive or -1 */
  if (hostBytesToPerform < -1)
    endProgram(12);

  /** Cycle through supplied options and parameters determining the sequential read/write methods and verifies correct arguments were supplied **/
  for (i=1; i<options.opts; i++)
  {
    /* are we being asked to reverse the order of any bits? */
    if (!strcmp(options.opt[i], "R"))
    {
      /* make sure supplied arguments are valid */
      if (char2int(options.arg[i][0]) < 1 || //argument 1 must be between 1 and 8
          char2int(options.arg[i][0]) > 8 ||
          char2int(options.arg[i][1]) < 1 || //same for argument 2
          char2int(options.arg[i][1]) > 8 ||
          options.arg[i][1] < options.arg[i][0] //don't need to be converted to integers, the comparison is still valid.
         )
        endProgram(9);
      continue; /* no need to check the rest */
    }

    /* are we being asked to invert any bits? */
    if (!strcmp(options.opt[i], "i"))
    {
      /* make sure supplied arguments are valid */
      for (c=0; c < strlen(options.arg[i]); c++)
        if (char2int(options.arg[i][c]) < 1 || char2int(options.arg[i][c]) > 8)
          endProgram(10);
      continue;
    }

    /* are we being asked to transpose any bits? */
    if (!strcmp(options.opt[i], "T"))
    {
      /* make sure supplied arguments are valid */
      if( (char2int(options.arg[i][0]) < 1 || char2int(options.arg[i][0]) > 8) || (char2int(options.arg[i][1]) < 1 || char2int(options.arg[i][1]) > 8))
        endProgram(11);
      continue;
    }
  }

  /** WRITE MODE **********/
  if (mode == 'w')
  {
    for (; head > 0; head--) /* skip host bytes if we were asked to */
    {
      if (fread(&hostByte, sizeof(uchar), 1, host)==0) /* read host bytes to skip them/make sure we're not at end of file */
        endProgram(1);

      /* we need to output this even though we don't write to it. HEAD OF HOST */
      writeHostByte();
    }

    if (fread(&hostByte, sizeof(uchar), 1, host)==0) /* read the first byte/make sure we're not at end of file */
      endProgram(1);

    while (1) /* main mode loop */
    {
      if (fread(&payByte, sizeof(uchar), 1, payload) == 0 || hostBytesPerformed == hostBytesToPerform || payloadBytesPerformed == payloadBytesToPerform) /* read payload byte/make sure we're not at end of file */
      {
        while (1) /* we've completed payload, so output the rest of the host file */
        {
          if (fread(&hostByte, sizeof(uchar), 1, host) == 0) /* if we're at end of host file */
            break; /* thats a bye bye */

          writeHostByte();
        }

        if (hostBytesPerformed == hostBytesToPerform)
          endProgram(13);

        if (payloadBytesPerformed == payloadBytesToPerform)
          endProgram(14);

        endProgram(0);
      }

      payByte=doBitManipFuncs(payByte); /* do sequential methods if supplied */

      /* pack LSB(s) into host */
      for (i=8; i>=1; i--)
      {
        if (bitsReadOrWritten == bitsPerHostByte)
        {
          writeHostByte();
          hostBytesPerformed++; /* thats one more host byte we've completed */
          if (hostBytesPerformed == hostBytesToPerform)
            break; /* break instead of ending program because we have to output the rest of the host bytes that are unmodified */

          bitsReadOrWritten=0; /* reset this because we've got a new byte to deal with */

          /* read a host byte, or if we are at end of file, exit */
          /* don't perform bytes in between byted that we're told not to, if we weren't then skip is 0 and this loop will be executed only once to read in the byte we need */
          for (hops=skip; hops>=0; hops--)
          {
            if (fread(&hostByte, sizeof(uchar), 1, host) == 0) /* read host byte */
              endProgram(1);

            if (hops > 0) /* if we're going to skip this byte, we still need to write it */
            {
              writeHostByte();
              hostBytesPerformed++; /* thats one more host byte we've completed, even though it was skipped */
              if (hostBytesPerformed == hostBytesToPerform)
                break; /* break instead of ending program because we have to output the rest of the host bytes that are unmodified */
            }
          }

          if (hostBytesPerformed == hostBytesToPerform)
            break; /* break instead of ending program because we have to output the rest of the host bytes that are unmodified */
        }

        /* bitsPerHostByte - bitsReadOrWritten will equate to the position in which the next bit should be written */
        hostByte=putBit(hostByte, getBit(payByte, i), bitsPerHostByte - bitsReadOrWritten); /* write it! */
        bitsReadOrWritten++; /* thats one more bit we've written */
      }

      payloadBytesPerformed++; /* and one more payload byte written */
    }
  }

  /** READ MODE **********/
  else if (mode=='r')
  {
    for (; head > 0; head--) /* skip bytes if neccessary */
      if (fread(&hostByte, sizeof(uchar), 1, host) == 0) /* read host byte/make sure we're not at end of file */
        endProgram(1);

    if (fread(&hostByte, sizeof(uchar), 1, host) == 0) /* read first byte/make sure we're not at end of file */
      endProgram(1);

    while(1) /* main mode loop */
    {
      /* pack those host LSB's into a byte */
      for (i=8; i>=1; i--)
      {
        if (bitsReadOrWritten == bitsPerHostByte) /* if we're done with this host byte, set up another */
        {
          bitsReadOrWritten=0;
          /* read a host byte, or if we are at end of file, exit */
          /* don't perform bytes in between byted that we're told not to, if we weren't then skip is 0 and this loop will be executed only once to read in the byte we need */
          for (hops=skip; hops>=0; hops--)
          {
            if (fread(&hostByte, sizeof(uchar), 1, host) == 0) /* read host byte */
              endProgram(1);

            hostBytesPerformed++; /* thats one more host byte we've completed, even though it was skipped or not */
            if (hostBytesPerformed == hostBytesToPerform)
              endProgram(13);
          }
        }

        /* bitsPerHostByte - bitsReadOrWritten will equate to the position in which the next bit should be written */
        payByte=putBit(payByte, getBit(hostByte, bitsPerHostByte - bitsReadOrWritten), i); /* write it! */

        bitsReadOrWritten++; /* thats one more bit we've written */
      }

      payByte=doBitManipFuncs(payByte); /* do sequential methods if supplied */

      writePayByte();
      payloadBytesPerformed++; /* and one more payload byte written */
      if (payloadBytesPerformed == payloadBytesToPerform)
        endProgram(14);

//      hostBytesPerformed++;    /* thats one more host byte read */
//      if (hostBytesPerformed == hostBytesToPerform)
//        endProgram(13);
    }
  }

  /** EVALUATE MODE **********/
  /* evaluate mode currently does not support hte following options: -h */
  else if (mode=='e')
  {
    printf("Host file can hold up to %i bytes using the supplied aguments.\n\n", ((((filelen(host)/8)-head)-tail)*bitsPerHostByte));
  }

  endProgram(0);
  return 0; /* unneccesary, but it feels wrong not to keep this here */
}

/** Cycle through supplied options and parameters determining the sequential read/write methods and performs them on a byte **/
uchar doBitManipFuncs(uchar byte)
{
  uint i, c;

  extern OPTIONS options;

  for (i=1; i<options.opts; i++)
  {
    if (!strcmp(options.opt[i], "R")) /* are we being asked to reverse the order of any bits? */
    {
      byte=reverseBits(byte, char2int(options.arg[i][0]), char2int(options.arg[i][1]));
    }
    else if (!strcmp(options.opt[i], "i")) /* are we being asked to invert any bits? */
    {
      /* for each bit to invert, do it! */
      for (c=0; c<strlen(options.arg[i]); c++)
      {
        byte=flipBit(byte, char2int(options.arg[i][c]));
      }
    }
    else if (!strcmp(options.opt[i], "T")) /* are we being asked to transpose any bits? */
    {
      byte=swapBits(byte, char2int(options.arg[i][0]), char2int(options.arg[i][1]));
    }
  }

  return byte;
}

/** free variables and exit **/
void endProgram(int value)
{
  int optIndex=-1;
  bool messagesOn=TRUE;

  extern FILE *host, *payload, *out;
  extern OPTIONS options;

  if (host != NULL)
  {
    fclose(host);
    host=NULL;
  }

  if (payload != NULL)
  {
    fclose(payload);
    payload=NULL;
  }

  if (out != NULL)
  {
    fclose(out);
    out=NULL;
  }

  /* are we being asked to suppress messages? */
  optIndex=getArg("m");
  if (optIndex != -1)
  {
    ucase(options.arg[optIndex]);
    if (strcmp(options.arg[optIndex], "OFF") == 0)
      messagesOn=FALSE;
  }

  if (messagesOn)
  {
    printf("\n");
    switch (value)
    {
      case 1:
        printf("Reached end of host file.\n");
        break;
      case 2:
        printf("Error opening host file for read access.\n");
        break;
      case 3:
        printf("Error opening host file for write access.\n");
        break;
      case 4:
        printf("Error opening payload file for read access.\n");
        break;
      case 5:
        printf("You must specify a payload file for write mode.\n");
        break;
      case 6:
        printf("You must specify an operation mode.\n");
        break;
      case 7:
        printf("Error opening output file for write access.\n");
        break;
      case 8:
        printf("Valid integers are 1-8 only for option '-b'.\n");
        break;
      case 9:
        printf("Option -R requires two integer arguments between 1 and 8, and the second must\n");
        printf("not be smaller than the first. Example: -R 37\n");
        break;
      case 10:
        printf("Option -i requires one to eight integer arguments between 1 and 8.\n");
        printf("Example: -i 12367\n");
        break;
      case 11:
        printf("Option -T requires two integer arguments between 1 and 8. Example: -t 15\n");
        break;
      case 12:
        printf("Option -B or -t require one positive integer argument greater than 0.\n");
        printf("Example: -B 150\n");
        break;
      case 13:
        printf("Host bytes to perform limit reached.\n");
        break;
      case 14:
        printf("Payload bytes to perform limit reached.\n");
        break;
    }
  }

  freeOptions();
  exit(value);
}
